@@ -2,7 +2,7 @@
 #
 #    Limited command Shell (lshell)
 #  
-#    $Id: lshell.py,v 1.80 2010/10/26 22:35:28 ghantoos Exp $
+#    $Id: lshell.py,v 1.80 2010-10-26 22:35:28 ghantoos Exp $
 #
 #    Copyright (C) 2008-2009 Ignace Mouzannar (ghantoos) <ghantoos@ghantoos.org>
 #
@@ -36,13 +36,13 @@ import re
 import getopt
 import logging
 import signal
-import subprocess
 import readline
 import grp
 import time
+import glob
 
-__author__ = "Ignace Mouzannar (ghantoos) <ghantoos@ghantoos.org>"
-__version__ = "0.9.14"
+__author__ = "Ignace Mouzannar <ghantoos@ghantoos.org>"
+__version__ = "0.9.15.1"
 
 # Required config variable list per user
 required_config = ['allowed', 'forbidden', 'warning_counter'] 
@@ -58,7 +58,7 @@ if sys.exec_prefix != '/usr':
 else:
     # for *Linux
     conf_prefix = ''
-config_file = conf_prefix + '/etc/lshell.conf'
+configfile = conf_prefix + '/etc/lshell.conf'
 
 # history file
 history_file = ".lhistory"
@@ -72,7 +72,7 @@ usage = """Usage: lshell [OPTIONS]
   --log    <dir>  : Log files directory
   -h, --help      : Show this help message
   --version       : Show version
-""" % config_file
+""" % configfile
 
 help_help = """Limited Shell (lshell) limited help.
 Cheers.
@@ -86,7 +86,7 @@ class ShellCmd(cmd.Cmd, object):
     """ Main lshell CLI class
     """
 
-    def __init__(self, userconf, stdin=None, stdout=None, stderr=None,         \
+    def __init__(self, userconf, args, stdin=None, stdout=None, stderr=None,         \
                                     g_cmd=None, g_line=None):
         if stdin is None:
             self.stdin = sys.stdin
@@ -101,6 +101,7 @@ class ShellCmd(cmd.Cmd, object):
         else:
             self.stderr = stderr
 
+        self.args = args
         self.conf = userconf
         self.log = self.conf['logpath']
 
@@ -112,7 +113,8 @@ class ShellCmd(cmd.Cmd, object):
         if self.conf.has_key('prompt'):
             self.promptbase = self.conf['prompt']
             self.promptbase = self.promptbase.replace('%u', getuser())
-            self.promptbase = self.promptbase.replace('%h', os.uname()[1])
+            self.promptbase = self.promptbase.replace('%h', \
+                                                    os.uname()[1].split('.')[0])
         else:
             self.promptbase = getuser()
 
@@ -120,6 +122,9 @@ class ShellCmd(cmd.Cmd, object):
 
         self.intro = self.conf['intro']
 
+        # initialize oldpwd variable to home directory
+        self.oldpwd = self.conf['home_path']
+
         # initialize cli variables
         self.g_cmd = g_cmd
         self.g_line = g_line
@@ -133,6 +138,13 @@ class ShellCmd(cmd.Cmd, object):
         the 'allowed' variable, and lshell will react as if you had            \
         added a do_uname in the ShellCmd class!
         """
+
+        # in case the configuration file has been modified, reload it
+        if self.conf['config_mtime'] != os.path.getmtime(self.conf['configfile']):
+            self.conf = CheckConfig(self.args).returnconf()
+            self.prompt = '%s:~$ ' % self.setprompt(self.conf)
+            self.log = self.conf['logpath']
+
         if self.g_cmd in ['quit', 'exit', 'EOF']:
             self.log.error('Exited')
             if self.g_cmd == 'EOF':
@@ -140,7 +152,7 @@ class ShellCmd(cmd.Cmd, object):
             sys.exit(0)
         if self.check_secure(self.g_line, self.conf['strict']) == 1: 
             return object.__getattribute__(self, attr)
-        if self.check_path(self.g_line) == 1:
+        if self.check_path(self.g_line, strict = self.conf['strict']) == 1:
             return object.__getattribute__(self, attr)
         if self.g_cmd in self.conf['allowed']:
             self.g_arg = re.sub('^~$|^~/', '%s/' %self.conf['home_path'],      \
@@ -172,6 +184,18 @@ class ShellCmd(cmd.Cmd, object):
         self.g_cmd, self.g_arg, self.g_line = ['', '', ''] 
         return object.__getattribute__(self, attr)
 
+    def setprompt(self, conf):
+        """ set prompt used by the shell
+        """
+        if conf.has_key('prompt'):
+            promptbase = conf['prompt']
+            promptbase = promptbase.replace('%u', getuser())
+            promptbase = promptbase.replace('%h', os.uname()[1].split('.')[0])
+        else:
+            promptbase = getuser()
+
+        return promptbase
+
     def lpath(self):
         """ lists allowed and forbidden path
         """
@@ -225,6 +249,28 @@ class ShellCmd(cmd.Cmd, object):
         """ implementation of the "cd" command
         """
         if len(self.g_arg) >= 1:
+            # add wildcard completion support to cd
+            if self.g_arg.find('*'):
+                # get all files and directories matching wildcard
+                wildall = glob.glob(self.g_arg)
+                wilddir = []
+                # filter to only directories
+                for item in wildall:
+                    if os.path.isdir(item):
+                        wilddir.append(item)
+                # sort results
+                wilddir.sort()
+                # if any results are returned, pick first one
+                if len(wilddir) >= 1:
+                    self.g_arg = wilddir[0]
+            # go previous directory
+            if self.g_arg == '-':
+                self.g_arg = self.oldpwd
+
+            # store current directory in oldpwd variable
+            self.oldpwd = os.getcwd()
+
+            # change directory
             try:
                 os.chdir(os.path.realpath(self.g_arg))
                 self.updateprompt(os.getcwd())
@@ -245,37 +291,56 @@ class ShellCmd(cmd.Cmd, object):
         is warned more than X time (X beeing the 'warning_counter' variable).
         """
 
+        if re.findall('[:cntrl:].*\n', line):
+            if not ssh:
+                if strict:
+                    self.counter_update('syntax')
+                else:
+                    self.log.critical('*** forbidden syntax -> %s' % line)
+            return 1
+
         for item in self.conf['forbidden']:
             # allow '&&' and '||' even if singles are forbidden
             if item in ['&', '|']:
                 if re.findall("[^\%s]\%s[^\%s]" %(item, item, item), line):
                     if not ssh:
-                        self.counter_update('syntax')
+                        if strict:
+                            self.counter_update('syntax')
+                        else:
+                            self.log.critical('*** forbidden syntax -> %s'    \
+                                                    % line)
                     return 1
             else:
                 if item in line:
                     if not ssh:
-                        self.counter_update('syntax')
+                        if strict:
+                            self.counter_update('syntax')
+                        else:
+                            self.log.critical('*** forbidden syntax -> %s'    \
+                                                    % line)
                     return 1
 
         returncode = 0
         # check if the line contains $(foo) executions, and check them
         executions = re.findall('\$\([^)]+[)]', line)
         for item in executions:
-            returncode += self.check_path(item[2:-1].strip())
-            returncode += self.check_secure(item[2:].strip(), strict=1)
+            returncode += self.check_path(item[2:-1].strip(), strict = strict)
+            returncode += self.check_secure(item[2:-1].strip(), strict = strict)
 
         # check fot executions using back quotes '`'
         executions = re.findall('\`[^`]+[`]', line)
         for item in executions:
-            returncode += self.check_secure(item[1:-1].strip(), strict=1)
+            returncode += self.check_secure(item[1:-1].strip(), strict = strict)
 
         # check if the line contains ${foo=bar}, and check them
         curly = re.findall('\$\{[^}]+[}]', line)
         for item in curly:
-            # split to get get variable only, and remove last character "}"
-            variable = re.split('=|\+|\?|\-', item, 1)
-            returncode += self.check_path(variable[1][:-1])
+            # split to get variable only, and remove last character "}"
+            if re.findall(r'=|\+|\?|\-', item):
+                variable = re.split('=|\+|\?|\-', item, 1)
+            else:
+                variable = item
+            returncode += self.check_path(variable[1][:-1], strict = strict)
             
         # if unknown commands where found, return 1 and don't execute the line
         if returncode > 0:
@@ -297,25 +362,26 @@ class ShellCmd(cmd.Cmd, object):
 
             # in case of a sudo command, check in sudo_commands list if allowed
             if command == 'sudo':
-                if cmdargs[1] not in self.conf['sudo_commands'] and cmdargs:
-                    if self.conf['strict'] == 1:
-                        if not ssh:
-                            self.counter_update('command')
-                    else:
-                        self.log.critical('*** forbidden sudo -> %s'   \
-                                                % line )
-                    return 1
+                if type(cmdargs) == list:
+                    if cmdargs[1] not in self.conf['sudo_commands'] and cmdargs:
+                        if self.conf['strict'] == 1:
+                            if not ssh:
+                                self.counter_update('command')
+                        else:
+                            self.log.critical('*** forbidden sudo -> %s'       \
+                                                    % line )
+                        return 1
             # if over SSH, replaced allowed list with the one of overssh
             if ssh:
                 self.conf['allowed'] = self.conf['overssh']
             
             # for all other commands check in allowed list
             if command not in self.conf['allowed'] and command:
-                if strict:
-                    if not ssh:
+                if not ssh:
+                    if strict:
                         self.counter_update('command', line)
-                else:
-                    self.log.critical('*** unknown command: %s' %command)
+                    else:
+                        self.log.critical('*** unknown command: %s' %command)
                 return 1
         return 0
          
@@ -346,22 +412,23 @@ class ShellCmd(cmd.Cmd, object):
                                     %(self.conf['warning_counter']))
                 self.stderr.write('This incident has been reported.\n')
 
-    def check_path(self, line, completion=None, ssh=None):
+    def check_path(self, line, completion=None, ssh=None, strict=None):
         """ Check if a path is entered in the line. If so, it checks if user   \
         are allowed to see this path. If user is not allowed, it calls         \
         self.counter_update. I case of completion, it only returns 0 or 1.
         """
-
         allowed_path_re = str(self.conf['path'][0])
         denied_path_re = str(self.conf['path'][1][:-1])
 
-        line = line.strip().split()
+        # split line depending on the operators
+        sep=re.compile(r'\ |;|\||&')
+        line = line.strip()
+        line = sep.split(line)
+
         for item in line:
             # remove potential quotes
-            try:
-                item = eval(item)
-            except:
-                pass
+            item = re.sub(r'^["\']|["\']$', '', item)
+
             # if item has been converted to somthing other than a string
             # or an int, reconvert it to a string
             if type(item) not in ['str', 'int']:
@@ -382,11 +449,15 @@ class ShellCmd(cmd.Cmd, object):
                 #    item = re.sub('\$%s|\${%s}' %(var, var), value, item)
                 # expand shell variables and wildcards using "echo"
                 # i know, this a bit nasty...
-                p = subprocess.Popen( "`which echo` %s" % item,
-                                      shell=True,
-                                      stdin=subprocess.PIPE,
-                                      stdout=subprocess.PIPE )
-                (cin, cout) = (p.stdin, p.stdout)
+                try:
+                    import subprocess
+                    p = subprocess.Popen( "`which echo` %s" % item,
+                                          shell=True,
+                                          stdin=subprocess.PIPE,
+                                          stdout=subprocess.PIPE )
+                    (cin, cout) = (p.stdin, p.stdout)
+                except ImportError:
+                    cin, cout = os.popen2('`which echo` %s' % item)
                 item = cout.readlines()[0].split(' ')[0].strip()
                 item = os.path.expandvars(item)
             tomatch = os.path.realpath(item)
@@ -398,15 +469,27 @@ class ShellCmd(cmd.Cmd, object):
             if not match_allowed or match_denied:
                 if not completion:
                     if not ssh:
-                        self.counter_update('path', tomatch)
-                return 1
+                        if strict: 
+                            self.counter_update('path', tomatch)
+                        else: 
+                            self.log.critical('*** Forbidden path: %s'        \
+                                                        % tomatch)
+                            return 1
+                    else:
+                        return 1
         if not completion:
             if not re.findall(allowed_path_re, os.getcwd()+'/'):
                 if not ssh:
-                    self.counter_update('path', os.getcwd())
-                    os.chdir(self.conf['home_path'])
-                    self.updateprompt(os.getcwd())
-                return 1
+                    if strict:
+                        self.counter_update('path', os.getcwd())
+                        os.chdir(self.conf['home_path'])
+                        self.updateprompt(os.getcwd())
+                    else:
+                        self.log.critical('*** Forbidden path: %s'            \
+                                                        %os.getcwd())
+                        return 1
+                else:
+                    return 1
         return 0
 
     def updateprompt(self, path):
@@ -415,7 +498,9 @@ class ShellCmd(cmd.Cmd, object):
 
         if path is self.conf['home_path']:
             self.prompt = '%s:~$ ' % self.promptbase
-        elif re.findall(self.conf['home_path'], path) :
+        elif self.conf['prompt_short'] == 1:
+            self.prompt = '%s: %s$ ' % (self.promptbase, path.split('/')[-1])
+        elif re.findall(self.conf['home_path'], path):
             self.prompt = '%s:~%s$ ' % ( self.promptbase, \
                                          path.split(self.conf['home_path'])[1])
         else:
@@ -443,8 +528,6 @@ class ShellCmd(cmd.Cmd, object):
             readline.set_completer(self.complete)
             readline.parse_and_bind(self.completekey+": complete")
         try:
-            if intro is not None:
-                self.intro = intro
             if self.conf['intro']:
                 self.stdout.write(str(self.conf['intro'])+"\n")
             stop = None
@@ -669,7 +752,9 @@ class CheckConfig:
 
         self.conf = {}
         self.conf, self.arguments = self.getoptions(args, self.conf)
-        self.check_file(self.conf['configfile'])
+        configfile = self.conf['configfile']
+        self.conf['config_mtime'] = self.get_config_mtime(configfile)
+        self.check_file(configfile)
         self.get_global()
         self.check_log()
         self.get_config()
@@ -689,8 +774,8 @@ class CheckConfig:
         #if '-c' not in arguments and '--config' not in arguments:
         #    usage()
 
-        # set config_file as default configuration file
-        conf['configfile'] = config_file
+        # set configfile as default configuration file
+        conf['configfile'] = configfile
 
         try:
             optlist, args = getopt.getopt(arguments,                           \
@@ -785,11 +870,20 @@ class CheckConfig:
             try:
                 logname = eval(self.conf['syslogname'])
             except:
-                logfilename = self.conf['syslogname']
+                logname = self.conf['syslogname']
         else:
             logname = 'lshell'
 
-        logger = logging.getLogger(logname)
+        logger = logging.getLogger("%s.%s" % (logname, \
+                                                self.conf['config_mtime']))
+
+        # close any logger handler/filters if exists
+        # this is useful if configuration is reloaded
+        for loghandler in logger.handlers:
+            logging.shutdown(logger.handlers)
+        for logfilter in logger.filters:
+            logger.removeFilter(logfilter)
+
         formatter = logging.Formatter('%%(asctime)s (%s): %%(message)s' \
                                                 % getuser() )
         syslogformatter = logging.Formatter('%s[%s]: %s: %%(message)s' \
@@ -878,9 +972,12 @@ class CheckConfig:
         grplist = os.getgroups()
         grplist.reverse()
         for gid in grplist:
-            grpname = grp.getgrgid(gid)[0]
-            section = 'grp:' + grpname
-            self.get_config_sub(section)
+            try:
+                grpname = grp.getgrgid(gid)[0]
+                section = 'grp:' + grpname
+                self.get_config_sub(section)
+            except KeyError:
+                pass
 
         # get user configuration if any
         self.get_config_sub(self.user)
@@ -908,7 +1005,8 @@ class CheckConfig:
                         elif stuff and key == 'path':
                             liste = ['', '']
                             for path in eval(stuff):
-                                liste[0] += os.path.realpath(path) + '/.*|'
+                                for item in glob.glob(path):
+                                    liste[0] += os.path.realpath(item) + '/.*|'
                             self.conf_raw.update({key:str(liste)})
                         elif stuff and type(eval(stuff)) is list:
                             self.conf_raw.update({key:stuff})
@@ -918,7 +1016,8 @@ class CheckConfig:
                 elif key == 'path':
                     liste = ['', '']
                     for path in self.myeval(value, 'path'):
-                        liste[0] += os.path.realpath(path) + '/.*|'
+                        for item in glob.glob(path):
+                            liste[0] += os.path.realpath(item) + '/.*|'
                     self.conf_raw.update({key:str(liste)})
                 else:
                     self.conf_raw.update(dict([item]))
@@ -1030,9 +1129,14 @@ class CheckConfig:
                     'strict',
                     'aliases',
                     'prompt',
+                    'prompt_short',
+                    'allowed_cmd_path',
                     'history_size']:
             try:
-                self.conf[item] = self.myeval(self.conf_raw[item], item)
+                if len(self.conf_raw[item]) == 0:
+                    self.conf[item] = ""
+                else:
+                    self.conf[item] = self.myeval(self.conf_raw[item], item)
             except KeyError:
                 if item in ['allowed', 'overssh', 'sudo_commands']:
                     self.conf[item] = []
@@ -1089,7 +1193,7 @@ class CheckConfig:
                 self.log.error('CONF: scpforce must be a string!')
 
         if self.conf_raw.has_key('intro'):
-            self.conf['intro'] = self.myeval(self.conf_raw['intro'])
+            self.conf['intro'] = self.conf_raw['intro']
         else:
             self.conf['intro'] = intro
 
@@ -1132,6 +1236,17 @@ class CheckConfig:
         if self.conf['sudo_commands']:
             self.conf['allowed'].append('sudo')
 
+        # add all commands present in allowed_cmd_path if specified
+        if self.conf['allowed_cmd_path']:
+            for path in self.conf['allowed_cmd_path']:
+                # add path to PATH env variable
+                os.environ['PATH'] += ":%s" % path
+                # find executable file, and add them to allowed commands
+                for item in os.listdir(path):
+                    cmd = os.path.join(path, item)
+                    if os.access(cmd, os.X_OK):
+                        self.conf['allowed'].append(item)
+
     def account_lock(self, user, lock_counter, check=None):
         """ check if user account is locked, in which case, exit """
         ### TODO ###
@@ -1165,21 +1280,9 @@ class CheckConfig:
                 # initialise cli session
                 cli = ShellCmd(self.conf, None, None, None, None,              \
                                                             self.conf['ssh'])
-                if cli.check_path(self.conf['ssh'], None, ssh=1):
+                if cli.check_path(self.conf['ssh'], None, ssh=1) == 1:
                     self.ssh_warn('path over SSH', self.conf['ssh'])
 
-                # check path first
-                allowed_path_re = str(self.conf['path'][0])
-                denied_path_re = str(self.conf['path'][1][:-1])
-                for item in self.conf['ssh'].strip().split(' '):
-                    tomatch = os.path.realpath(item) + '/'
-                    match_allowed = re.findall(allowed_path_re, tomatch)
-                    if denied_path_re:
-                        match_denied = re.findall(denied_path_re, tomatch)
-                    else: match_denied = None
-                    if not match_allowed or match_denied:
-                        self.ssh_warn('path over SSH', self.conf['ssh'])
-
                 # check if scp is requested and allowed
                 if self.conf['ssh'].startswith('scp '):
                     if self.conf['scp'] is 1 or 'scp' in self.conf['overssh']: 
@@ -1281,6 +1384,13 @@ class CheckConfig:
                 sys.exit(0)
         else: return 0
 
+    def get_config_mtime(self, configfile):
+        """ get configuration file modification time, and store in the        \
+            configuration dict. This should then be used to reload the 
+            configuration dynamically upon file changes
+        """
+        return os.path.getmtime(configfile)
+
     def returnconf(self):
         """ returns the configuration dict """
         return self.conf
@@ -1297,17 +1407,29 @@ class LshellTimeOut(Exception):
 def get_aliases(line, aliases):
     """ Replace all configured aliases in the line
     """
+
     for item in aliases.keys():
-        reg = '(^|; |;)%s([ ;&\|]+|$)' % item
-        while re.findall(reg, line):
-            beforecommand = re.findall(reg, line)[0][0]
-            aftercommand = re.findall(reg, line)[0][1]
-            line = re.sub(reg, "%s%s%s" % (beforecommand, aliases[item],       \
-                                                     aftercommand), line, 1)
-            # if line does not change after sub, exit loop
+        reg1 = '(^|; |;)%s([ ;&\|]+|$)(.*)' % item
+        reg2 = '(^|; |;)%s([ ;&\|]+|$)' % item
+
+        # in case aliase bigin with the same command
+        # (this is until i find a proper regex solution..)
+        if aliases[item].startswith(item):
+            aliaskey = os.urandom(10)
+
+        while re.findall(reg1, line):
+            (before, after, rest) = re.findall(reg1, line)[0]
             linesave = line
+            cmd = "%s %s" % (item, rest)
+
+            line = re.sub(reg2, "%s%s%s" % (before, aliaskey,       \
+                                                     after), line, 1)
+            # if line does not change after sub, exit loop
             if linesave == line:
                 break
+        # replace the key by the actual alias
+        line = line.replace(aliaskey, aliases[item])
+
     for char in [';', '&', '|']:
         # remove all remaining double char
         line = line.replace('%s%s' %(char, char), '%s' %char)
@@ -1324,7 +1446,7 @@ def main():
     userconf = CheckConfig(args).returnconf()
 
     try:
-        cli = ShellCmd(userconf)
+        cli = ShellCmd(userconf, args)
         cli.cmdloop()
 
     except (KeyboardInterrupt, EOFError):
